import inspect
import logging
import time


logger = logging.getLogger(__name__)


class Tag(object):
    """``class`` that abstract tags and act as a simple container."""
    def __init__(self, name, valuable, ttl=-1):
        """
        :param string name: tag name
        :param valuable: something that gives tag value, string or function
        :param None,int ttl: Time to live for caching the tags, possible values
            are:
                * None (means no ttl)
                * -1 (means infinite ttl)
                * positive integer in seconds
        """
        self.name = name
        self._value = valuable
        self.ttl = ttl

    @property
    def value(self):
        """Property that return tag value. Just return the provided value or
        call the function.

        """
        if inspect.isfunction(self._value):
            try:
                return self._value()
            except Exception:
                logger.exception(u'Calculating tag %s failed:' % self.name)
                return None

        return self._value


class DynamicTags(object):
    """Set tags dynamicaly.

    Tags must be defined in the constructor.

    Example use:

        * first subclass :class:`DynamicTags`:
            .. code-block:: python

                class Disks(DynamicTags):
                    def __init__(self):
                        # find disk by calling an arbitrary function
                        disks = find_disks()

                        # for each disk associate a function that calculate
                        # the size
                        for d in disks:
                            # arbitrary example: for each disk set a tag
                            # with the disk name and the size as value with
                            # a ttl of one hour
                            self.tags.append(
                                Tag(d, find_disk_size(d), 3600),
                            )

                        # we're done :)
        * we now just need to instanciate our new ``class``:
            .. code-block:: python

                disks_size = Disks()

        * the tags are automaticaly found using :func:`tag_inspector`
    """
    def __init__(self):
        #: iterable containing :class:`Tag` objects (``list`` by default)
        self.tags = []

    def iter_tags(self):
        for t in self.tags:
            yield t


def tag_inspector(mod):
    """Inspect module to find tags.

    :param module mod: module to inspect

    Currently there are three ways to define a tag inside a module:

        * affect a string to a variable, the variable name will be the tag
          name
        * define a function that returns a value as a string or None (meaning
          the tag doesn't exist on the host), as you guessed the function name
          will define the tag name
        * subclass :class:`DynamicTags` and create an instance. The instance
          must have an attribute ``tags`` which must be an **iterable** of
          :class:`Tag` objects (by default it is a ``list``). The name of the
          instance isn't meaningfull.
    """
    tags = []
    for n, m in inspect.getmembers(mod):  # (name, member)
        # only keep strings or functions as tags
        if getattr(m, '__module__', None) != mod.__name__ or (
            n.startswith('_')):
            continue
        elif isinstance(m, (str, unicode)):
            # if string, it means it is constant, then ttl = -1
            ttl = -1
        elif inspect.isfunction(m):
            # if function take function ttl argument or set -1 by default
            ttl = getattr(m, 'ttl', -1)
        elif isinstance(m, DynamicTags):
            for t in m.iter_tags():
                tags.append(t)
                logger.debug('Introspected %s with ttl %s.' % (t.name,
                                                               t.ttl))
            continue
        else:
            # whatever it is we don't care...
            continue

        logger.debug('Introspected %s with ttl %s.' % (n, ttl))

        # finally add the tag
        tags.append(Tag(n, m, ttl))

    return tags


def get_tags(obj, tags, noresolve_tags):
    """Helper to get tags.

    :param obj: object with a dict ``tags`` attribute
    :param tags: List of tags to get (None mean all tags)
    :param noresolve_tags: List of tags to exclude from result
    """
    # TODO clean this fucntion
    logger.debug('Tags request: %s, %s' % (
        unicode(tags),
        unicode(noresolve_tags),
    ))

    tags = set(tags) - set(noresolve_tags) if tags is not None else None

    if tags is None:
        tags = obj.tags.iterkeys()
    else:
        tags = tags & set(obj.tags)

    profile = time.time()
    try:
        result = dict((
            t,  # tag name
            dict(
                value=obj.tags[t].value,
                ttl=obj.tags[t].ttl,
            ),
        ) for t in tags)
    except Exception:
        logger.exception(u'SHould not happend, result.')

    logger.debug(u'Profiling: %f seconds.' % (time.time() - profile))
    logger.debug(u'Returning: %s' % unicode(result))
    return result
